module("L_HoneywellLib_legacy", package.seeall)

--------------------------------------------------------------------------------
--
-- Honeywell Total Connect Comfort backend for legacy API.
--
-- Creator:  Mikee from JoeyD's original plugin

-- Version 1.0: Dec 27, 2015 Mikee
--   Initial
--
-- Revised by rigpapa 2020-02-14 LuaSec 0.8 requires "protocol" be set in requests ("any" is usually good)

local main = { }

-- Upcalls
local trace = nil
local log = nil
local err = nil
local log_http = nil
local dump = nil
local url_encode = nil
local inject = nil
local trim = nil
local task = nil
local clearTask = nil
local updateCookies = nil
local createCookie = nil

local json = require("L_HoneywellTCC_json")
local https = require("ssl.https")
local K = require("L_HoneywellTCC_const")

--------------------------------------------------------------------------------
-- various constants
--
local USER_AGENT = "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1500.95 Safari/537.36"
local TCC_SID = "urn:joeyd-com:serviceId:HoneywellTCC1"

--------------------------------------------------------------------------------
-- various globals
--
local TccDevice -- Control Device ID
local TblLoginCookie = nil -- table of cookies for refresh authentication.
local ForceTccStatus = 0
local ForceSetStatus = 0

--------------------------------------------------------------------------------
-- constants for login and request credentials
--
local HOST = "mytotalconnectcomfort.com"
local URL_PORTAL = "https://" .. HOST .. "/portal/"
local URL_DEVICE_STATUS = URL_PORTAL .. "Device/CheckDataSession/"
local URL_SET_DEVICE_STATUS = URL_PORTAL .. "Device/SubmitControlScreenChanges"
local URL_LOCATIONS = URL_PORTAL .. "Locations"

--------------------------------------------------------------------------------
--

--++
-- Connects to the Honeywell Total Connect Comfort (TCC) API
-- Returns result and resultText, result is boolean, true if  is successful,
-- false otherwise, resultText holds the reason why a failure occured OR the
-- required cookie
--
-- Many thanks to mikee and and CudaNet for their help in getting authentication,
-- to get this working!
--
-- @param username - username
-- @param password - password
-- @return result - K.STATUS_OK (success), K.STATUS_xxx (fail)
--         resultText - string reason for status
--	   theCookie - login cookie
--
-- local
local function retrieveLoginCookie(username, password)
   trace("retrieveLoginCookie", { username = username, password = password } );

   local result
   local resultText
   local theCookie = {} 
   local auth_response = {}
   local theBody = "timeOffset=300&UserName=" .. url_encode(username) .. "&Password=" ..
      url_encode(password) .. "&RememberMe=false"

   -- response is the page, status= a value (see https://docs.python.org/2/library/httplib.html) 
   -- header contains the reponse authentication header table 
   local response, status, header = https.request{
      url = URL_PORTAL,
      method = "POST",
      headers = {
	 ["Content-Type"] = "application/x-www-form-urlencoded",
	 ["Accept"] = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
	 ["Accept-Encoding"] = "sdch",
	 ["Host"] = HOST,
	 ["DNT"] = "1",
	 ["Origin"] = URL_PORTAL,
	 ["User-Agent"] = USER_AGENT,
	 ["Content-Length"] = #theBody,
      },
      source = ltn12.source.string(theBody),
      sink = ltn12.sink.table(auth_response),
	  protocol = 'any'
   }

   if ForceTccStatus ~= 0 then
      status = ForceTccStatus
   end

   log_http("auth_response" .. dump(auth_response))
   log_http("RESPONSE:" .. dump(response)) 
   log_http("STATUS:" .. dump(status))
   log_http("HEADER:" .. dump(header))

   if tostring(status) == "302" then
      result = K.STATUS_OK
      resultText = "Authorization Successful. Cookie obtained."

      theCookie = updateCookies(header["set-cookie"],theCookie)
      log_http("theCookie " .. dump(theCookie))
   else
      result = K.STATUS_AUTH_FAIL
      resultText = "Authentication Error.  Error code: " .. tostring(status)

      theCookie = {}
   end

   trace("retrieveLoginCookie return", { result = result,
					 resultText = resultText,
					 theCookie = theCookie })
   return result, resultText, theCookie
end

--++
-- read the location info which includes the location name and location id.
--
-- @param htmlPage - HTML page with location info embedded
-- @return table of { locationId { Name = name, Devices = {} } } (note device table is empty)
--
-- local
local function getLocationIDsFromHTML(htmlPage)
   trace("getLocationIDsFromHTML") -- , { htmlPage = htmlPage })

   local theResults = {}  -- table with the index, location number, and value the location name
   for location,_ in string.gmatch(htmlPage,'url="/portal/(%d+)/Zones"') do
      local theLoc = string.find(htmlPage, location)
      local name = string.match(htmlPage, '<div class="location%-name">(.-)</div>', theLoc)
      name = trim(name)
      theResults[tonumber(location)] = { Name = tostring(name), Devices = {} }
   end

   trace("getLocationIDsFromHtml return", { theResults = theResults })
   return theResults
end

--+
-- Get the thermostat devices at the given TCC location
--
-- @param locNumber - TCC location number
-- @return resultCode - K.STATUS_OK (success), K.STATUS_xxx (fail)
--         resultText - string reason for result code
--         tblDevices - table of devices
--
-- local
local function getDevicesInLocation(locNumber)
   trace("getDevicesInLocation", { locNumber = locNumber })

   local resultCode = K.STATUS_FAIL
   local resultText = ""
   local webPage_table = {}	
   local webPage = ""
   local tblDevices = { }

   log("ABOUT TO MAKE HTTPS CALL FOR ZONES")
   local response, status, header = https.request{
      url = "https://mytotalconnectcomfort.com/portal/" .. tostring(locNumber) .. "/Zones",
      method = "GET",
      headers = {
	 ["Accept"] = "*/*",
	 ["DNT"] = "1",
	 ["Accept-Encoding"] = "plain",
	 ["Accept-Language"] = "en-US,en;q=0.8",
	 ["Connection"] = "keep-alive",
	 ["Host"] = HOST,
	 ["Referer"]="https://mytotalconnectcomfort.com/portal",
	 ["User-Agent"] = USER_AGENT,
	 ["Cookie"] = createCookie(TblLoginCookie)
      },
      sink = ltn12.sink.table(webPage_table),
	  protocol = 'any'
   }

   if ForceTccStatus ~= 0 then
      status = ForceTccStatus
   end

   log_http("webPage_table" .. dump(webPage_table))
   log_http("RESPONSE:" .. dump(response)) 
   log_http("STATUS:" .. dump(status))
   log_http("HEADER:" .. dump(header))

   if tostring(status)=="200" then
      webPage = table.concat(webPage_table) -- contains the entire web page source
      webPage = inject("GetDevices", webPage, "zones_page.html")

      -- Update the cookie
      TblLoginCookie = updateCookies(header["set-cookie"], TblLoginCookie)
      -- parse out the device IDs from within the HTML of the web page

      local deviceIDs = { }
      local deviceIndex = 1;
      for device,_ in string.gmatch(webPage, 'data%-id="(%d+)"') do
	 deviceIDs[deviceIndex] = device;
	 deviceIndex = deviceIndex + 1;
      end

      local deviceNames = { }
      local nameIndex = 1;
      for name,_ in string.gmatch(webPage, 'class="location%-name">([%a ]+)') do
	 deviceNames[nameIndex] = name;
	 nameIndex = nameIndex + 1;
      end

      if deviceIndex == nameIndex then
	 for i=1,deviceIndex-1 do
	    tblDevices[deviceIDs[i]] = deviceNames[i];
	 end
      end

      local numLocs = 1;
      if numLocs == 0 then
	 resultCode = K.STATUS_FAIL
	 resultText = "No locations found when getting device information using Zones"
      elseif deviceIndex == 1 then
	 resultCode = K.STATUS_FAIL
	 resultText = "Location found but did not return any devices when using Zones"
      else
	 resultCode = K.STATUS_OK
	 resultText = "Success"
      end

   elseif tostring(status)=="401" then
      resultCode = K.STATUS_AUTH_FAIL
      resultText = "Unauthorized request when trying to get devices with Zones"
   else
      resultCode = K.STATUS_HTTP_FAIL
      resultText = "Unknown error when trying to get devices with Zones.  HTTP Code " .. tostring(status)
   end
   if resultCode ~= K.STATUS_OK then
      log(resultText)
   end

   trace("getDevicesInLocation return", { resultCode, resultText, tblDevices })
   return resultCode, resultText, tblDevices
end

--++
-- this function will attempt to obtain the list of TCC locations by making
-- HTTP calls that returns a web page. Once the locations are known a
-- subsequent API call to "<location>/Zones" is made for each location to
-- construct a resulting table of locations and devices. (no publicly known
-- API calls for getting the locations is known at this time)
--
-- @return resultCode - K.STATUS_OK (success), K.STATUS_xxx (fail)
--         resultText - string reason for result code
--         tblLocations - table with an entry for each location containing a list of the devices at that location
--
-- local
local function retrieveLocationsAndDevices()
   trace("retrieveLocationsAndDevice");

   local resultCode = K.STATUS_FAIL
   local resultText = ""
   local webPage_sink = {}	
   local webPage = ""

   log("ABOUT TO MAKE HTTPS CALL FOR LOCATIONS")

   local response, status, header = https.request{
      url = URL_LOCATIONS,
      method = "GET",
      headers = {
	 ["Accept"] = "*/*",
	 ["DNT"] = "1",
	 ["Accept-Encoding"] = "plain",
	 ["Accept-Language"] = "en-US,en;q=0.8",
	 ["Connection"] = "keep-alive",
	 ["Host"] = HOST,
	 ["Referer"]="https://mytotalconnectcomfort.com/portal",
	 ["User-Agent"] = USER_AGENT,
	 ["Cookie"] = createCookie(TblLoginCookie)
      },
      sink = ltn12.sink.table(webPage_sink),
	  protocol = 'any'
   }

   if ForceTccStatus ~= 0 then
      status = ForceTccStatus
   end

   log_http("webPage_sink" .. dump(webPage_sink))
   log_http("RESPONSE:" .. dump(response)) 
   log_http("STATUS:" .. dump(status))
   log_http("HEADER:" .. dump(header))

   local tblLocations = {}

   if tostring(status)=="200" then
      webPage = table.concat(webPage_sink)  -- contains the entire web page source
      webPage = inject("GetLocations", webPage, "loc_page.html")

      -- Update the cookie
      TblLoginCookie = updateCookies(header["set-cookie"], TblLoginCookie)

      -- parse out the location IDs from within the HTML of the web page
      tblLocations = getLocationIDsFromHTML(webPage)
      if not next(tblLocations) then
	 -- no locations were found
	 resultCode = K.STATUS_FAIL
	 resultText = "Successful https POST to get locations, but no locations found."

      else
	 -- now we can make API calls to add the device thermostat information
	 for i,v in pairs(tblLocations) do
	    local devResult, decText, theDevice = getDevicesInLocation(i)
	    if devResult == K.STATUS_OK then
	       for j,k in pairs(theDevice) do
		  tblLocations[i].Devices[j] = {Name = k}
	       end
	       resultCode = K.STATUS_OK -- successful retrieval of location and device info
	       resultText = "Success"
	    else
	       resultCode = K.STATUS_REMOTE_FAIL
	       resultText = "Issue with processing getDevicesInLocation.  (Issue already logged)"
	    end
	 end
      end
   elseif tostring(status)=="401" then
      resultCode = K.STATUS_AUTH_FAIL
      resultText = "Unauthorized request."
   else
      resultCode = K.STATUS_HTTP_FAIL
      resultText = "Unsuccessful post request to honeywell locations web page. HTTP Code " .. tostring(status)
   end
   if resultCode ~= K.STATUS_OK then
      log(resultText)
   end

   trace("retrieveLocationsAndDevice return", { resultCode = resultCode,
						resultText = resultText,
						tblLocations = tblLocations })
   return resultCode, resultText, tblLocations
end

--++
-- Function to convert the json status to the generic library status
--
-- @param theData - json status
-- @return generic status record
--
local function convertStatus(theData)
   trace("convertStatus", { theData=theData })
   local rec = { }
   if theData ~= nil then
      rec.Success = theData.success
      rec.Load = theData.latestData.drData.Load
      rec.FanSwitchPosition = theData.latestData.fanData.fanMode
      rec.HasFan = theData.latestData.hasFan
      rec.CoolNextPeriod = theData.latestData.uiData.CoolNextPeriod
      rec.CoolSetpoint = theData.latestData.uiData.CoolSetpoint
      rec.DisplayUnits = theData.latestData.uiData.DisplayUnits
      rec.DisplayTemp = theData.latestData.uiData.DispTemperature
      rec.HeatNextPeriod = theData.latestData.uiData.HeatNextPeriod
      rec.HeatSetpoint = theData.latestData.uiData.HeatSetpoint
      rec.IndoorHumidity = theData.latestData.uiData.IndoorHumidity
      rec.HasIndoorHumidity = theData.latestData.uiData.IndoorHumiditySensorAvailable
      rec.OutdoorHumidity = theData.latestData.uiData.OutdoorHumidity
      rec.HasOutdoorHumidity = theData.latestData.uiData.OutdoorHumiditySensorAvailable
      rec.OutdoorTemp = theData.latestData.uiData.OutdoorTemperature
      rec.HasOutdoorTemp = theData.latestData.uiData.OutdoorTemperatureAvailable
      rec.SchedCoolSetpoint = theData.latestData.uiData.ScheduleCoolSp
      rec.SchedHeatSetpoint = theData.latestData.uiData.ScheduleHeatSp
      rec.StatusCool = theData.latestData.uiData.StatusCool
      rec.StatusHeat = theData.latestData.uiData.StatusHeat
      rec.CanSwitchAuto = theData.latestData.uiData.SwitchAutoAllowed
      rec.SystemSwitchPosition = theData.latestData.uiData.SystemSwitchPosition
   end
   trace("convertStatus return", { rec=rec })
   return rec
end

--++
-- Get the status of a single thermostat
--
-- @param deviceId - TCC device ID
-- @return resultCode - K.STATUS_OK (success), K.STATUS_xxx (fail)
--         resultText - string indicating reason for code
--         currentData - current data from TCC
--
-- local
local function getStatus(deviceId)
   trace("getStatus", { deviceId = deviceId })

   local currentData = {}
   local resultCode = K.STATUS_FAIL
   local resultText = ""
   local status_sink = {}	
	
   local response, status, header = https.request{
      url = URL_DEVICE_STATUS .. deviceId,
      method = "GET",
      headers = {
	 ["Accept"] = "*/*",
	 ["DNT"] = "1",
	 ["Accept-Encoding"] = "plain",
	 ["Cache-Control"] = "max-age=0",
	 ["Accept-Language"] = "en-US,en,q=0.8",
	 ["Connection"] = "keep-alive",
	 ["Host"] = HOST,
	 ["Referer"] = "https://mytotalconnectcomfort.com/portal",
	 ["X-Requested-With"] = "XMLHttpRequest",
	 ["User-Agent"] = USER_AGENT,
	 ["Cookie"] = createCookie(TblLoginCookie)
      },
      sink = ltn12.sink.table(status_sink),
	  protocol = 'any'
   }

   if ForceTccStatus ~= 0 then
      status = ForceTccStatus
   end

   log_http("status_sink" .. dump(status_sink))
   log_http("RESPONSE:" .. dump(response)) 
   log_http("STATUS:" .. dump(status))
   log_http("HEADER:" .. dump(header))

   log_http("COOKIE STRING: " .. createCookie(TblLoginCookie))

   if tostring(status)=="200" then

      -- Update the cookie
      TblLoginCookie = updateCookies(header["set-cookie"], TblLoginCookie)
      local status_str = table.concat(status_sink)
      status_str = inject("GetStatus", status_str, "status.json")

      log_http("status_str", { status_str = status_str })

      status_str = string.gsub(status_str,'%[%s-%]','[null]') 
      status_str = string.gsub(status_str,'[\r\n]','') 
      log_http("status_str after gsub", { status_str = status_str })

      currentData = convertStatus(json.decode(status_str))
      log_http("current_data", { currentData = currentData })

      if currentData == nil then
	 resultCode = K.STATUS_FAIL
	 resultText = "Request accepted, empty response"
      else
	 if tostring(currentData.Success) == "true" then
	    resultCode = K.STATUS_OK
	    resultText = "Request successfully processed."
	 else
	    resultCode = K.STATUS_FAIL
	    resultText = "Request accepted, but not successfully processed."
	 end
      end

   elseif tostring(status)=="401" then
      resultCode = K.STATUS_AUTH_FAIL
      resultText = "Unauthorized request."

   else
      resultCode = K.STATUS_HTTP_FAIL
      resultText = "Unknown error.  HTTP Code " .. tostring(status)
   end

   trace("getStatus return", { resultCode = resultCode,
				    resultText = resultText,
				    currentData = currentData })
   return resultCode, resultText , currentData
end

--++
-- Send the thermostat values to TCC
--
-- @param req - set request table (see buildSetReq)
-- @return result - K.STATUS_OK (success), K.STATUS_xxx (fail)
--         resultText - string reason for result code
--
-- local
local function setStatus(req)
   trace("setStatus", { req = req })

   -- req.DeviceID = the numeric honeywell ID
	
   local result
   local resultText
   local screenChange_sink = {}

   local body = json.encode(req)
   local res, status, header = https.request {
      method = "POST",
      ["url"] = URL_PORTAL .. "Device/SubmitControlScreenChanges",
      headers = {
         ["accept"] = "application/json, text/javascript, *.*; q=0.01",
         ["accept-encoding"] = "gzip,deflate,sdch",
         ["accept-language"] = "en-US, en, q=0.8",
         ["cache-control"] = "max-age=0",
         ["connection"] = "keep-alive",
         ["host"] = HOST,
         ["dnt"] = "1",
         ["referer"] = URL_PORTAL .. "Device/Control/" .. tostring(req.DeviceID),
         ["x-requested-with"] = "XMLHttpRequest",
         ["user-agent"] = USER_AGENT,
         ["content-type"] = "application/json",
         ["content-length"] = #body,
         ["cookie"] = createCookie(TblLoginCookie)
      },
      source = ltn12.source.string(body),
      sink = ltn12.sink.table(screenChange_sink),
	  protocol = 'any'
   }
   
   if ForceTccStatus ~= 0 then
      status = ForceTccStatus
   end
   if ForceSetStatus ~= 0 then
      status = ForceSetStatus
   end

   log_http("screenChange_sink" .. dump(screenChange_sink))
   log_http("RESPONSE:" .. dump(res)) 
   log_http("STATUS:" .. dump(status))
   log_http("HEADER:" .. dump(header))

   if tostring(status) == "200" then
      -- successful put		
      local response = table.concat(screenChange_sink)
      local screenChange = json.decode(response)	
      if tostring(screenChange.success) == "1" then
	 -- verified success by honewell API
	 result = K.STATUS_OK
	 resultText = "Command processed successfully."
      else
	 -- set command was accepted by API, but honeywell returned unsuccessful action.
	 result = K.STATUS_FAIL
	 resultText = "Command posted, but not processed successfully by Honeywell."
      end
   elseif tostring(status) == "401" then
      -- Request came back unauthorized
      result = K.STATUS_AUTH_FAIL
      resultText = "Unauthorized request."
   else	
      -- some other code
      result = K.STATUS_HTTP_FAIL
      resultText = "Unknown error.  HTTP Code " .. tostring(status)
   end
   
   trace("setStatus return", { result = result, resultText = resultText })
   return result, resultText
end

--------------------------------------------------------------------------------
--

--++
-- Perform initialization
--
function Init(devId, imports)
   TccDevice = devId
   main = imports

   trace = main.trace
   log = main.log
   log_http = main.log_http
   dump = main.dump
   err = main.err

   url_encode = main.url_encode
   inject = main.inject
   trim = main.trim

   task = main.task
   clearTask = main.clearTask

   updateCookies = main.updateCookies
   createCookie = main.createCookie
end

--++
-- Login and return whatever token is reuired for subsequent calls. This token
-- will be passed on all library calls
--
-- @param user - string username
-- @param pass - string password
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         token - token to use on subsequent calls
function Login(user, pass)
   trace("Login", { user=user, pass=pass })

   local status, reason, token = retrieveLoginCookie(user, pass)

   trace("Login return", { status=status, reason=reason })
   return status, reason, token
end

--++
-- Return a structure of location/device information
--
-- @param token - authentication token
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         devices {
--                  { locationID = {
--                    Name = "",  - string location name
--                    Devices = {
--                     { deviceId = name },...
--                    }
--                   },...
--                  }
function GetDevices(token)
   trace("GetDevices", { token=token })

   TblLoginCookie = token
   if TblLoginCookie == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason, devices = retrieveLocationsAndDevices()

   trace("GetDevices return", { status=status, reason=reason })
   return status, reason, devices
end

--++
-- Return the status of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         status - see L_HoneywellTCC_const.lua for detail of GetStatus
--
function GetStatus(token, deviceId)
   trace("GetStatus", { token=token })

   TblLoginCookie = token
   if TblLoginCookie == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason, data = getStatus(deviceId)

   trace("GetStatus return", { status=status, reason=reason })
   return status, reason, data
end

--++
-- Set the operating parameters of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @param req - see L_HoneywellTCC_const.lua for detail of SetStatus
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--
function SetStatus(token, deviceId, req)
   trace("SetStatus", { token=token })

   TblLoginCookie = token
   if TblLoginCookie == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   req.DeviceID = tonumber(req.DeviceID)
   local status, reason = setStatus(req)

   trace("SetStatus return", { status=status, reason=reason })
   return status, reason
end
