module("L_HoneywellLib_mobile", package.seeall)

--------------------------------------------------------------------------------
--
-- Honeywell Total Connect Comfort backend for MobileV2 API.
--
-- Creator:  Mikee

-- Version 1.0: Dec 27, 2015 Mikee
--   Initial

local main = { }

-- Upcalls
local debug = nil
local trace = nil
local log = nil
local err = nil
local log_http = nil
local url_encode = nil
local lua_to_url = nil
local xml_to_lua = nil

local https = require("ssl.https")
local K = require("L_HoneywellTCC_const")

--------------------------------------------------------------------------------
-- Constants

-- DeviceID = number
local SESSION_ID = "SessionID"
local THERMOSTAT_ID = "ThermostatID"

-- Honeywell Web portal constants
--local K_HOST = "rs.alarmnet.com"
--local K_URL = "https://" .. K_HOST .. "/TotalConnectComfort/ws/MobileV2.asmx/"
local K_HOST = "tccna.honeywell.com"
local K_URL = "https://" .. K_HOST .. "/ws/MobileV2.asmx/"
local K_USER_AGENT = "Apache-HttpClient/UNAVAILABLE (java 1.4)"

--------------------------------------------------------------------------------
-- Globals
local Session = nil
local TccDevice = 0
local RetryLimit = 1 -- no retries

--------------------------------------------------------------------------------
-- Local functions

--++
-- Make an http request with retry
--
-- @param spec_options - table of additional or override options
-- @return res - 
--         code - https result code
--         header - response headers
--
function https_request(spec_options)
   trace("https_request", { url = spec_options.url })

   local res, code, header 
   local base_options = {
      ["headers"] = {
	 ["Accept"] = "*/*",
--	 ["Accept-Encoding"] = "gzip,deflate,sdch",
	 ["Accept-Encoding"] = "sdch",
	 ["Accept-Language"] = "en-US",
	 ["Cache-Control"] = "max-age=0",
	 ["Connection"] = "keep-alive",
	 ["user-agent"] = K_USER_AGENT,
	 ["Host"] = K_HOST,
      },
   }

   for i=1,RetryLimit do
      local options = { }
      options.headers = { }
      for name,value in pairs(base_options) do
	 if name == "headers" then
	    for hname, hvalue in pairs(value) do
	       options.headers[hname] = hvalue
	    end
	 else
	    options[name] = value
	 end
      end
      for name,value in pairs(spec_options) do
	 if name == "headers" then
	    for hname, hvalue in pairs(value) do
	       options.headers[hname] = hvalue
	    end
	 else
	    options[name] = value
	 end
      end

      -- Fix for LuaSec 0.8 and up, rigpapa
      options.protocol = 'any'

      log_http("https req", { options=options })
      res, code, header = https.request(options)
      log_http("https resp", { res=res, code=code, header=header })

      if code ~= 400 and code ~= 504 then
	 break
      end
      trace("https_request retry", { res=res, code=code })
   end

   trace("https_request return", { res=res, code=code })
   return res, code, header
end

--++
-- Login to the Honeywell Total Connect Comfort website
--
-- @param user - username
-- @param pass - password
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         token - session token
-- ------------------------------------------------------------------------------
local function login(user, pass)
   local time = os.time();

   local args = {
      ["ApplicationID"] = "660d663b-53f4-43c2-872c-bf3a4357b5fd",
      ["Username"] = user,
      ["UiLanguage"] = "English",
      ["ApplicationVersion"] = 2,
      ["Password"] = pass,
   }

   local url = K_URL .. "AuthenticateUserLogin"
   debug("url = " .. url)

   local body = lua_to_url(args)
   debug("body = " .. body)

   local result = { }
   local res, code, header = https_request {
      ["method"] = "POST",
      ["url"] = url,
      ["headers"] = {
	 ["content-type"] = "application/x-www-form-urlencoded",
	 ["content-length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   last_login_attempt = os.time();

   if (code ~= 200) then
      reason = "Authentication failed, code = " .. code
      status = K.STATUS_FAIL
      return status, reason, nil
   end

   debug("result = " .. table.concat(result))
   local res = xml_to_lua(table.concat(result))
   result = res.AuthenticateLoginResult.Result
   if (result ~= "Success") then
      reason = "Authentication failed, result = " .. (result or "nil")
      status = K.STATUS_FAIL
      return status, reason, nil
   end

   Session = res.AuthenticateLoginResult.SessionID

   return K.STATUS_OK, "Success", Session
end

--++
-- Get the list of devices per location
--
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         locations - table of locations (See GetLocations for description)
--
local function get_locations()

   local args = {
      [SESSION_ID] = Session,
   }

   local url = K_URL .. "GetLocations"
   debug("url = " .. url)

   local body = lua_to_url(args)
   debug("body = " .. body)

   local result = { }
   local res, code, header = https_request {
      ["method"] = "POST",
      ["url"] = url,
      ["headers"] = {
	 ["content-type"] = "application/x-www-form-urlencoded",
	 ["content-length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   if (code ~= 200) then
      reason = "GetLocations failed, code = " .. code
      status = K.STATUS_FAIL
      return status, reason, nil
   end

   debug("result = " .. table.concat(result))
   res = xml_to_lua(table.concat(result))
   log("res", { res = res })

   result = res.GetLocationsResult.Result
   if (result ~= "Success") then
      reason = "Authentication failed, result = " .. (result or "nil")
      status = K.STATUS_FAIL
      return status, reason, nil
   end

   local devices = { }
   local locations = res.GetLocationsResult.Locations
   for loc=0,32 do
      local location = nil
      if (loc == 0) then
	 location = locations["LocationInfo"]
      else
	 location = locations["LocationInfo_" .. loc]
      end
      if (location == nil) then
	 break;
      end

      local thermostats = res.GetLocationsResult.Locations.LocationInfo.Thermostats

      local tbl = { }
      for i=0,32 do
	 local thermo = nil
	 if (i == 0) then
	    thermo = thermostats["ThermostatInfo"]
	 else
	    thermo = thermostats["ThermostatInfo_" .. i]
	 end
	 if (thermo == nil) then
	    break;
	 end

	 -- ["id"] = thermo.ThermostatID,
	 -- ["name"] = thermo.DeviceName,
	 -- ["mac"] = thermo.MacID,
	 -- ["user_name"] = thermo.UserDefinedDeviceName,

	 tbl[thermo.ThermostatID] = thermo.DeviceName
      end
      devices[location.LocationID] = { }
      devices[location.LocationID].Devices = tbl
      devices[location.LocationID].Name = location.Name
   end

   return K.STATUS_OK, "Success", devices
end

--++
-- Get the status of a thermostat
--
-- @param device_id = TCC device id
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--         stat - table of thermostat data (see GetStatus for description)
--
local function get_status(device_id)

   local args = {
      [SESSION_ID] = Session,
      [THERMOSTAT_ID] = device_id,
   }

   local url = K_URL .. "GetThermostat"
   debug("url = " .. url)

   local body = lua_to_url(args)
   debug("body = " .. body)

   local result = { }
   local res, code, header = https_request {
      ["method"] = "POST",
      ["url"] = url,
      ["headers"] = {
	 ["content-type"] = "application/x-www-form-urlencoded",
	 ["content-length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   if (code ~= 200) then
      local reason = "GetStatus failed, code = " .. code
      local status = K.STATUS_FAIL
      return status, reason, nil
   end

   debug("result = " .. table.concat(result))
   local res = xml_to_lua(table.concat(result))

   local result = res.GetThermostatResult.Result
   if (result ~= "Success") then
      local reason = "GetStatus failed, code = " .. code
      local status = K.STATUS_FAIL
      return status, reason, nil
   end

   result = res.GetThermostatResult.Thermostat

   res = { }
   res.Success = true
   res.Load = 0

   res.FanSwitchPosition = K.FAN_MODE_MAP1[result.Fan.Position]

   res.HasFan = tostring(result.Fan.CanControl) == "true"
   res.CoolNextPeriod = tonumber(result.UI.CoolNextPeriod) or 0
   res.CoolSetpoint = tonumber(result.UI.CoolSetpoint) or 0
   res.DisplayUnits = result.UI.DisplayedUnits
   res.DisplayTemp = tonumber(result.UI.DispTemperature) or 0
   res.HeatNextPeriod = tonumber(result.UI.HeatNextPeriod) or 0
   res.HeatSetpoint = tonumber(result.UI.HeatSetpoint) or 0

   res.IndoorHumidity = tonumber(result.UI.IndoorHumidity) or 128
   res.HasIndoorHumidity = tonumber(result.UI.IndoorHumidity) ~= 128

   res.OutdoorHumidity = tonumber(result.UI.OutdoorHumidity) or 128
   res.HasOutdoorHumidity = tonumber(result.UI.OutdoorHumidity) ~= 128

   res.OutdoorTemp = tonumber(result.UI.OutdoorTemp) or 128
   res.HasOutdoorTemp = tonumber(result.UI.OutdoorTemp) ~= 128

   res.SchedCoolSetpoint = tonumber(result.UI.SchedCoolSp) or 0
   res.SchedHeatSetpoint = tonumber(result.UI.SchedHeatSp) or 0
   res.StatusCool = tonumber(result.UI.StatusCool) or 0
   res.StatusHeat = tonumber(result.UI.StatusHeat) or 0
   res.CanSwitchAuto = tostring(result.UI.CanSetSwitchAuto) == "true"
   res.SystemSwitchPosition = tonumber(result.UI.SystemSwitchPosition) or 0

   return K.STATUS_OK, "Success", res
end

--++
-- Obtain schedule from the website for specified thermostat
--
-- @param device_id - TCC device id
--
local function get_schedule(device_id)

   local args = {
      [SESSION_ID] = Session,
      [THERMOSTAT_ID] = device_id,
   }

   local url = K_URL .. "GetSchedule"
   debug("url = " .. url)

   local body = lua_to_url(args)
   debug("body = " .. body)

   local result = { }
   local res, code, header = https_request {
      ["method"] = "POST",
      ["url"] = url,
      ["headers"] = {
	 ["content-type"] = "application/x-www-form-urlencoded",
	 ["content-length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   if (code ~= 200) then
      local reason = "GetSchedule failed, code = " .. code
      local status = K.STATUS_FAIL
      return status, reason, nil
   end

   debug("result = " .. table.concat(result))
   res = xml_to_lua(table.concat(result))

   result = res.ScheduleResult.Result
   if (result ~= "Success") then
      local reason = "GetSchedule failed, code = " .. code
      local status = K.STATUS_FAIL
      return status, reason, nil
   end

   return K.STATUS_OK, "Success", nil
end

function setit(url, body)
   trace("setit", { url=url, body=body} )

   local result = { }
   local res, code, header = https_request {
      ["method"] = "POST",
      ["url"] = url,
      ["headers"] = {
	 ["content-type"] = "application/x-www-form-urlencoded",
	 ["content-length"] = #body,
      },
      ["source"] = ltn12.source.string(body),
      ["sink"] = ltn12.sink.table(result),
   }

   local status, reason
   if (code ~= 200) then
      reason = "SetStatus failed, code = " .. code
      if (code == 401) then
	 status = K.STATUS_AUTH_FAIL
      else
	 status = K.STATUS_FAIL
      end
   else
      debug("result = " .. table.concat(result))
      res = xml_to_lua(table.concat(result))

      result = res.ChangeRequestResult.Result
      if (result ~= "Success") then
	 reason = "SetStatus failed, code = " .. code
	 status = K.STATUS_FAIL
      else
	 status = K.STATUS_OK
	 reason = "Success"
      end
   end

   trace("set_status return", { status=status, reason=reason })
   return status, reason
end

--++
-- Set thermostat values
--
-- @param device_id - TCC device ID
-- @param req - Requested state change (See SetStatus for description
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - string reason for above code
--
local function set_status(device_id, req)
   trace("set_status", { device_id=device_id, req=req })

   local url = K_URL .. "ChangeThermostatUI"

   local args = { }
   args[SESSION_ID] = Session
   args[THERMOSTAT_ID] = device_id

   -- If there is a system mode change make sure to execute it first
   if (req[K.SYSTEM_SWITCH] ~= nil) then

      args[K.SYSTEM_SWITCH] = req[K.SYSTEM_SWITCH]
      args["Change" .. K.SYSTEM_SWITCH] = true

      args[K.HEAT_SETPOINT] = 0
      args["Change" .. K.HEAT_SETPOINT] = false
      args[K.STATUS_HEAT] = 0
      args["Change" .. K.STATUS_HEAT] = false
      args[K.HEAT_NEXT] = 0
      args["Change" .. K.HEAT_NEXT] = false

      args[K.COOL_SETPOINT] = 0
      args["Change" .. K.COOL_SETPOINT] = false
      args[K.STATUS_COOL] = 0
      args["Change" .. K.STATUS_COOL] = false
      args[K.COOL_NEXT] = 0
      args["Change" .. K.COOL_NEXT] = false

      local body = lua_to_url(args)
      local status, reason = setit(url, body)
      if status ~= K.STATUS_OK then
	 return status, reason
      end
   end

   args[K.SYSTEM_SWITCH] = 0
   args["Change" .. K.SYSTEM_SWITCH] = false

   if (req[K.HEAT_SETPOINT] ~= nil) then
      args["Change" .. K.HEAT_SETPOINT] = true
      args[K.HEAT_SETPOINT] = req[K.HEAT_SETPOINT]
      args[K.STATUS_HEAT] = req[K.STATUS_HEAT]
      if (args[K.STATUS_HEAT] == nil) then
	 args[K.STATUS_HEAT] = K.STATUS_TYPE_TEMP -- XXX use prop behave
      end
      args["Change" .. K.STATUS_HEAT] = true
   else
      args[K.HEAT_SETPOINT] = 0
      args["Change" .. K.HEAT_SETPOINT] = false
      args[K.STATUS_HEAT] = 0
      args["Change" .. K.STATUS_HEAT] = false
   end

   args[K.HEAT_NEXT] = 0
   args["Change" .. K.HEAT_NEXT] = false

   if (req[K.COOL_SETPOINT] ~= nil) then
      args["Change" .. K.COOL_SETPOINT] = true
      args[K.COOL_SETPOINT] = req[K.COOL_SETPOINT]
      args[K.STATUS_COOL] = req[K.STATUS_COOL]
      if (args[K.STATUS_COOL] == nil) then
	 args[K.STATUS_COOL] = K.STATUS_TYPE_TEMP -- XXX use prop behave
      end
      args["Change" .. K.STATUS_COOL] = true
   else
      args[K.COOL_SETPOINT] = 0
      args["Change" .. K.COOL_SETPOINT] = false
      args[K.STATUS_COOL] = 0
      args["Change" .. K.STATUS_COOL] = false
   end

   args[K.COOL_NEXT] = 0
   args["Change" .. K.COOL_NEXT] = false

   local body = lua_to_url(args)
   local status, reason = setit(url, body)
   return status, reason
end

--------------------------------------------------------------------------------
-- APIs

--++
-- Perform initialization
--
function Init(devId, imports)
   TccDevice = devId
   main = imports

   debug = main.log_http
   trace = main.trace
   log = main.log
   log_http = main.log_http
   err = main.err

   url_encode = main.url_encode

   lua_to_url = main.lua_to_url
   xml_to_lua = main.xml_to_lua
end

--++
-- Login and return whatever token is reuired for subsequent calls. This token
-- will be passed on all library calls
--
-- @param user - string username
-- @param pass - string password
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         token - token to use on subsequent calls
function Login(user, pass)
   trace("Mobile.Login", { user=user, pass=pass })

   local status, reason, session = login(user, pass)

   trace("Mobile.Login return", { status=status, reason=reason, session=session })
   return status, reason, session
end

--++
-- Return a structure of location/device information
--
-- @param token - authentication token
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         devices {
--                  { locationID = {
--                    Name = "",  - string location name
--                    Devices = {
--                     { deviceId = name },...
--                    }
--                   },...
--                  }
function GetDevices(token)
   trace("Mobile.GetDevices", { token=token })

   Session = token
   if Session == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason, devices = get_locations()

   trace("Mobile.GetDevices return", { status=status, reason=reason, devices=devices })
   return status, reason, devices
end

--++
-- Return the status of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--         status - see L_HoneywellTCC_const for details of GetStatus record
--
function GetStatus(token, deviceId)
   trace("Mobile.GetStatus", { token=token, deviceId=deviceId })

   Session = token
   if Session == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason, info = get_status(deviceId)

   trace("Mobile.GetStatus return", { status=status, reason=reason, info=info })
   return status, reason, info
end

--++
-- Set the operating parameters of a thermostat
--
-- @param token - authentication token
-- @param deviceId - device id
-- @param req - see L_HoneywellTCC_const for SetStatus record
-- @return status - K.STATUS_OK, K.STATUS_xxx
--         reason - reason for status above
--
function SetStatus(token, deviceId, req)
   trace("Mobile.SetStatus", { token=token, deviceId=deviceId, req=req })

   Session = token
   if Session == nil then
      err("Invalid token")
      return K.STATUS_AUTH_FAIL, "Invalid token", nil
   end

   local status, reason = set_status(deviceId, req)

   trace("Mobile.SetStatus return", { status=status, reason=reason })
   return status, reason
end
